/********************************************************************************
* Copyright (c) Erik Kunze 1995 - 1999
*
* Permission to use, distribute, and sell this software and its documentation
* for any purpose is hereby granted without fee, provided that the above
* copyright notice appear in all copies and that both that copyright notice and
* this permission notice appear in supporting documentation, and that the name
* of the copyright holder not be used in advertising or publicity pertaining to
* distribution of the software without specific, written prior permission.  The
* copyright holder makes no representations about the suitability of this
* software for any purpose.  It is provided "as is" without express or implied
* warranty. THE CODE MAY NOT BE MODIFIED OR REUSED WITHOUT PERMISSION!
*
* THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
* INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
* EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
* CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
* DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
* TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
* PERFORMANCE OF THIS SOFTWARE.
*
* Author: Erik Kunze
*******************************************************************************/
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#ifdef XZX_SCORPION
#ifndef lint
static char rcsid[] = "$Id: scorpion.c,v 1.7 1999/05/22 15:03:23 erik Rel $";
#endif
#include "z80.h"
#include "debug.h"
#include "resource.h"
#include "mem.h"
#include "io.h"
#include "machine.h"
#include "util.h"
#ifdef AUDIO
#include "audio.h"
#endif
#include "emul.h"
#ifdef JOY
#include "joystick.h"
#endif
#include "keyboard.h"
#ifdef XZX_KMOUSE
#include "kmouse.h"
#endif
#include "loadsave.h"
#include "screen.h"
#include "tables.h"
#include "trdos.h"
#include "scorpion.h"
#define P_BANKX			0x1ffd
#define B_BANKX			0xe002
#define B_SPECIALX		0x01
#define B_SELROMX		0x02
#define B_SELRAMX		0x10
#define EXTENDED_MEM(port) \
((port) & (B_SPECIALX | B_SELROMX))
#ifdef DEBUG
#define DEB(x)			{ if (GETCFG(debug) & D_IO) { x } }
#else
#define DEB(x)
#endif
static void scorReset(void);
static void scorNmi(void);
static uns8 scorReadOpcode(uns16);
static uns8 scorReadMem(uns16);
static void scorWriteMem(uns16, uns8);
static void scorInitPages(void);
static void scorPageIn(int, int);
static inline void scorOutBankm128(uns8);
static inline void scorOutBankmX(uns8);
static void scorOutPort(uns16, uns8);
static uns8 scorInPort(uns16);
void
ScorInit(void)
{
MachineSettings *msp;
int checksum, romtype, i, j, k;
msp = &Machines[SP_SCORPION];
msp->name = "Scorpion";
msp->basicRom = SCORROM1;
msp->romType = ROM_UNKNOWN;
msp->tvLines = 312;
msp->tstatesPerLine = 224;
msp->tstatesPerFrame = 69888;
msp->intLine = 0;
msp->topBorder = 64;
msp->ulaDelay = Malloc(msp->tstatesPerFrame, "ScorInit");
msp->reset = scorReset;
msp->nmi = scorNmi;
msp->readOpcode = scorReadOpcode;
msp->readMem = scorReadMem;
msp->writeMem = scorWriteMem;
msp->inPort = scorInPort;
msp->outPort = scorOutPort;
msp->initPages = scorInitPages;
msp->pageIn = scorPageIn;
if (!LoadRomImage(GETCFG(scorRom), SCORROM0, 4))
{
checksum = GetRomChecksum(msp->basicRom, XZX_PAGESIZE);
switch (checksum)
{
case 22821:
romtype = ROM_SCORPION;
break;
default:
Msg(M_WARN, "unknown Scorpion ROM image - checksum %u",
checksum);
romtype = ROM_UNKNOWN;
break;
}
msp->romType = romtype;
#ifndef NO_ED_TRAPS
if (romtype != ROM_UNKNOWN)
{
RealMemory[msp->basicRom][0x0562] = 0xed;
RealMemory[msp->basicRom][0x0563] = LOAD_BPT;
RealMemory[msp->basicRom][0x04c2] = 0xed;
RealMemory[msp->basicRom][0x04c3] = SAVE_BPT;
#ifdef XBELL_AUDIO
RealMemory[msp->basicRom][0x03b5] = 0xed;
RealMemory[msp->basicRom][0x03b6] = BELL_BPT;
#endif
}
#endif
}
for (i = 0; i < 14335; i++)
{
msp->ulaDelay[i] = 0x80;
}
for (j = 192; j; j --)
{
for (k = 128; k; k--)
{
msp->ulaDelay[i++] = 0;
}
for (k = 96; k; k--)
{
msp->ulaDelay[i++] = 0x80;
}
}
for (; i < msp->tstatesPerFrame; i++)
{
msp->ulaDelay[i] = 0x80;
}
}
static void
scorReset(void)
{
Z80_Reset();
Machine->initPages();
Last0x7FFD = 0;
ScreenSelect = RAM5;
Last0x1FFD = 0;
KbdReset();
(void)memset(InPorts, 0x00, sizeof(InPorts));
(void)memset(OutPorts, 0x00, sizeof(OutPorts));
Last0xFFFD = 0;
(void)memset(Psg, 0x00, sizeof(Psg));
#ifdef AYCHIP_AUDIO
AyReset();
#endif
}
static void
scorNmi(void)
{
Z80_NMI();
(void)Machine->inPort(SCORROM3);
}
static uns8
scorReadOpcode(uns16 addr)
{
uns8 op;
if (addr >= 0x4000)
{
if (RPAGE(0) == SCORROM3)
{
Machine->pageIn(0, Machine->basicRom);
}
}
else if ((addr & 0xff00) == 0x3d00)
{
if (RPAGE(0) == Machine->basicRom)
{
Machine->pageIn(0, SCORROM3);
}
}
op = MEM(addr);
return op;
}
static uns8
scorReadMem(uns16 addr)
{
uns8 byte;
byte = MEM(addr);
return byte;
}
static void
scorWriteMem(uns16 addr, uns8 byte)
{
int page = VPAGE(addr);
if (RealPageNumbers[page] >= RAM0)
{
TheMemory[page][addr & 0x1fff] = byte;
if (RealPageNumbers[page] == ScreenSelect)
{
if ((addr &= 0x3fff) < (uns16)PIXEL_LENGTH)
{
ScreenWrite(addr, byte);
}
else if (addr < (uns16)(PIXEL_LENGTH + ATTR_LENGTH))
{
AttrWrite(addr, byte);
}
}
}
}
static void
scorInitPages(void)
{
Machine->pageIn(0, SCORROM0);
Machine->pageIn(1, RAM5);
Machine->pageIn(2, RAM2);
Machine->pageIn(3, RAM0);
}
static void
scorPageIn(int z80page, int vpage)
{
int i = z80page << 1;
TheMemory[i] = RealMemory[vpage];
RealPageNumbers[i] = vpage;
i++;
TheMemory[i] = RealMemory[vpage] + XZX_PAGESIZE / 2;
RealPageNumbers[i] = vpage;
}
static inline void
scorOutBankm128(uns8 val)
{
DEB(Msg(M_DEBUG,
"bankm: RAM select %x, ROM select %d, screen %d, paging %sabled",
((Last0x1FFD & B_SELRAMX) >> 1) | (val & B_SELRAM),
(val & B_SELROM) >> 4, (val & B_SELSCREEN) >> 3,
PAGING_DISABLED(Last0x7FFD) ? "dis" : "en"););
if (PAGING_DISABLED(Last0x7FFD))
{
return;
}
ScreenSelect = val & B_SELSCREEN ? RAM7 : RAM5;
if ((val & B_SELSCREEN) != (Last0x7FFD & B_SELSCREEN))
{
ForceScreenRefresh();
}
Last0x7FFD = val;
Machine->pageIn(3, RAM0 +
(((Last0x1FFD & B_SELRAMX) >> 1) | (val & B_SELRAM)));
if (!EXTENDED_MEM(Last0x1FFD))
{
Machine->pageIn(0, SCORROM0 + ((val & B_SELROM) >> 4));
}
}
static inline void
scorOutBankmX(uns8 val)
{
int page0;
DEB(Msg(M_DEBUG,
"bankx: RAM select %x, RAM8 %d, ROM2 %d",
((val & B_SELRAMX) >> 1) | (Last0x7FFD & B_SELRAM),
val & B_SPECIALX, (val & B_SELROMX) >> 1););
if (PAGING_DISABLED(Last0x7FFD))
{
return;
}
switch (EXTENDED_MEM(val))
{
default:
page0 = SCORROM0 + ((Last0x7FFD & B_SELROM) >> 4);
break;
case B_SPECIALX:
page0 = RAM8;
break;
case B_SPECIALX | B_SELROMX:
Msg(M_WARN, "RAM and ROM together at page 0 is impossible");
case B_SELROMX:
page0 = SCORROM2;
break;
}
Machine->pageIn(0, page0);
Machine->pageIn(3, RAM0 +
(((val & B_SELRAMX) >> 1) | (Last0x7FFD & B_SELRAM)));
Last0x1FFD = val;
}
static void
scorOutPort(uns16 port, uns8 val)
{
DEB(Msg(M_DEBUG, "writing <%02x> to port <%04x>", val, port););
if (RPAGE(0) >= SCORROM2)
{
TrdosOutByte(port, val);
}
if ((port & B_ULA) == (P_ULA & B_ULA))
{
if (LastBorderColor != (val & B_BORDER))
{
LastBorderColor = val & B_BORDER;
SetBorderColor(LastBorderColor);
}
#ifdef SPEAKER_AUDIO
SpeakerOutByte(val);
#endif
#ifdef EMULATE_ULA_DELAY
#endif
}
if ((port & B_128) == (P_128 & B_128))
{
switch (port & 0xc000)
{
case P_SNDCONTROL & 0xc000:
Last0xFFFD = val;
break;
case P_SNDDATA & 0xc000:
if (Last0xFFFD < NELEM(Psg))
{
Psg[Last0xFFFD] = val;
if (Last0xFFFD == 14)
{
DEB(Msg(M_DEBUG, "writing AY register 14"););
}
#ifdef AYCHIP_AUDIO
else if (Last0xFFFD < 14)
{
AyOutByte(Last0xFFFD, val);
}
#endif
}
else
{
DEB(Msg(M_DEBUG, "writing AY register %d", Last0xFFFD););
}
break;
case P_BANK128 & 0xc000:
scorOutBankm128(val);
break;
default:
switch (port & 0x3000)
{
case P_BANKX & 0x3000:
scorOutBankmX(val);
break;
}
}
}
OutPorts[LOWBYTE(port)] = val;
}
static uns8
scorInPort(uns16 port)
{
uns8 res;
if (RPAGE(0) >= SCORROM2
&& (port & 0x001f) == 0x001f
&& ((port & 0x0080) == 0x0000 || (port & 0x00f0) == 0x00f0))
{
res = TrdosInByte(port);
goto quit;
}
#ifdef XZX_KMOUSE
INBYTE_KMOUSE(port, res, quit);
#endif
if ((port & B_KEMPSTON) == (P_KEMPSTON & B_KEMPSTON))
{
#ifdef JOY
if (GETCFG(joyActive) > 0)
{
InPorts[P_KEMPSTON] = (uns8)JoyRead();
}
#endif
res = InPorts[P_KEMPSTON] &
(B_RIGHT | B_LEFT | B_DOWN | B_UP | B_FIRE);
goto quit;
}
if ((port & B_ULA) == (P_ULA & B_ULA))
{
res = 0x1f;
if (!(port & 0x8000))
{
res &= KeyPorts[7];
}
if (!(port & 0x4000))
{
res &= KeyPorts[6];
}
if (!(port & 0x2000))
{
res &= KeyPorts[5];
}
if (!(port & 0x1000))
{
res &= KeyPorts[4];
}
if (!(port & 0x0800))
{
res &= KeyPorts[3];
}
if (!(port & 0x0400))
{
res &= KeyPorts[2];
}
if (!(port & 0x0200))
{
res &= KeyPorts[1];
}
if (!(port & 0x0100))
{
res &= KeyPorts[0];
}
res ^= TpGetNextBit();
#ifdef EMULATE_ULA_DELAY
#endif
goto quit;
}
if ((port & B_128) == (P_128 & B_128))
{
switch (port & 0xc000)
{
case P_SNDCONTROL & 0xc000:
if (Last0xFFFD == 14)
{
DEB(Msg(M_DEBUG, "reading AY register 14"););
res = Psg[7] & 0x40 ? (Psg[14] & 0x1f) | 0x20 : 0x3f;
}
else if (Last0xFFFD < 14)
{
res = (Psg[Last0xFFFD] & PsgMask[Last0xFFFD]);
}
else if (Last0xFFFD == 15)
{
res = (Psg[7] & 0x80 ? Psg[15] : 0xff);
}
else
{
res = 0xff;
}
goto quit;
case P_SNDDATA & 0xc000:
res = 0xff;
goto quit;
}
}
if (!(Machine->ulaDelay[TSTATES % Machine->tstatesPerFrame] & 0x80))
{
unsigned int tstatesInLine = TSTATES % Machine->tstatesPerLine;
unsigned int line = Vline - Machine->topBorder;
res = (RealMemory[ScreenSelect][(tstatesInLine & 0x02 ?
Line2Attr[line] : Line2Pix[line]) +
(tstatesInLine >> 2)]);
}
else
{
res = 0xff;
}
quit:
DEB(Msg(M_DEBUG, "reading <%02x> from port <%04x>", res, port););
return res;
}
#endif
